local Teletype = {}
Teletype.__index = Teletype

local maxLen = 405
local maxHeight = 676
local paperScrollSpeed = 43
local charsPerRow = 50

function Teletype:updateWaiting(dt)
  if #self.lineQueue ~= 0 then
    local l = self.lineQueue[1]
    table.remove(self.lineQueue, 1)
    self:startPrinting(l)
  end
end


function Teletype:clearBuffer()
  for i = 2, #self.lineQueue do
    self.lineQueue[i] = nil
  end
end


function Teletype:startPrinting(line)
  table.insert(self.lineBuf, 1, line[1])
  table.remove(self.lineBuf, 40)
  self.currentState = "Printing"

  local oldC = love.graphics.getCanvas()
  love.graphics.setCanvas(self.paperCanvas)
  love.graphics.clear(255, 255, 255, 255)
  love.graphics.setColor(64,64,64,255)
  love.graphics.setFont(G.teletypeFont)
  for i, l in ipairs(self.lineBuf) do
    love.graphics.print(l, 0, maxHeight - i * 20)
  end
  love.graphics.setCanvas(oldC)
  local id = self.paperCanvas:newImageData(0,0,maxLen, maxHeight)
  self.paperImage = love.graphics.newImage(id, {mipmaps=true})
  self.paperImage:setFilter('linear', 'linear', 10)
  --self.paperImage:setMipmapFilter('linear', 10)

  love.graphics.setColor(255,255,255,255)
  self.scroll = 20

  self.currentLength = #line[1] * 20
  self.finishFunc = line[2]
  G.sounds.teletype:play()
end

function Teletype:initCanvas()
  local oldC = love.graphics.getCanvas()
  love.graphics.setCanvas(self.paperCanvas)
  love.graphics.clear(255, 255, 255, 255)
  love.graphics.setCanvas(oldC)
  local id = self.paperCanvas:newImageData(0,0,maxLen, maxHeight)
  self.paperImage = love.graphics.newImage(id, {mipmaps=true})
  self.paperImage:setFilter('linear', 'linear', 10)
  --self.paperImage:setMipmapFilter('linear', 10)
end

function Teletype:updatePrinting(dt)
  if self.currentHeadPos < self.currentLength then
    self.currentHeadPos = self.currentHeadPos + 3000 * dt
  end

  if self.currentHeadPos >= self.currentLength then
    self.currentHeadPos = self.currentLength
    self.currentState = "Returning"
    G.sounds.teletype:stop()
    G.sounds.teletype2:play()
  end
end


function Teletype:updateReturning(dt)
  if self.currentHeadPos > 0 then
    self.currentHeadPos = self.currentHeadPos - 3000 * dt
    if self.currentHeadPos < 0 then
      self.currentHeadPos = 0
    end
  end

  if self.scroll > 0 then
    self.paperScroll = self.paperScroll + dt * paperScrollSpeed

    self.scroll = self.scroll - dt * 100
    if self.scroll <= 0 then
      self.scroll = 0
    end
  end

  if self.currentHeadPos == 0 and self.scroll == 0 then
    if self.finishFunc then
      self.finishFunc()
    end
    self.currentState = "Waiting"
    G.sounds.teletype2:stop()
  end
end


function Teletype:update(dt)
  self["update" .. self.currentState](self, dt)
end

function Teletype:draw()
  love.graphics.setColor(255,255,255,255)
  
  G.overlays.teletype:draw(math.floor(self.currentHeadPos / charsPerRow))
  G.overlays.paperHolesRight:draw(math.floor(self.paperScroll) % 9)
  G.overlays.paperHolesLeft:draw(math.floor(self.paperScroll) % 9)

  love.graphics.stencil(function()
    --love.graphics.rectangle('fill', 0, 0, 1300, 700)
    love.graphics.setShader(G.maskShader)
    G.overlays.teletype_stencil:draw(math.floor(self.currentHeadPos / charsPerRow))
  end)
  love.graphics.setStencilTest("less", 1)
  --love.graphics.setShader()
  --love.graphics.setColor(255,255,255,127)
  --G.overlays.teletype_stencil:draw(math.floor(self.currentHeadPos / 40))
  --love.graphics.setColor(255,255,255,255)


  love.graphics.setShader(G.paperShader)
  G.paperShader:send("paperContent", self.paperImage)
  G.paperShader:send("scroll", self.scroll / (maxHeight))
  love.graphics.setBlendMode('multiply')
  love.graphics.draw(G.paperMesh)
  love.graphics.setBlendMode('alpha')
  love.graphics.setShader()
  love.graphics.setStencilTest()

end


function Teletype:print(line, func)
  for i = 1, #line,charsPerRow do
    self.lineQueue[#self.lineQueue+1] = {line:sub(i,i+charsPerRow-1)}
  end

  self.lineQueue[#self.lineQueue][2] = func
end


function newTeletype()
  local new = {
    paperCanvas = love.graphics.newCanvas(maxLen*2, maxHeight),
    currentHeadPos = 0,
    lineBuf = {},
    lineQueue = {},
    scroll = 0,
    currentState = "Waiting",
    paperScroll = 0
  }

  setmetatable(new, Teletype)

  new:initCanvas()

  return new
end


return newTeletype
